import base64
import json


def base64url_decode(input_bytes: bytes) -> bytes:
    rem = len(input_bytes) % 4

    if rem > 0:
        input_bytes += b"=" * (4 - rem)

    return base64.urlsafe_b64decode(input_bytes)


def parse_jwt_for_expiration(string: str) -> int:
    jwt = string.encode("utf-8")

    signing_input, _ = jwt.rsplit(b".", 1)
    _, payload_segment = signing_input.split(b".", 1)

    payload = base64url_decode(payload_segment)
    payload_data = json.loads(payload)

    assert isinstance(payload_data["exp"], int)

    return payload_data["exp"]



from typing import Tuple, List, Optional, Iterator, Union, Literal, overload

import re
from time import sleep
from ural import is_url
from urllib.parse import urljoin
from urllib3 import PoolManager

from minet.web import request, create_pool_manager, Response
from minet.scrape import WonderfulSoup, Tag

from minet.reddit.constants import OLD_REDDIT_BASE_URL
from minet.reddit.exceptions import RedditInvalidTargetError
from minet.reddit.types import (
    RedditPost,
    RedditComment,
    RedditUserPost,
    RedditUserComment,
)


ID_RE = re.compile(r"t1_(\w+)")
COMMENTS_NUMBER_RE = re.compile(r"(\d+)\s+comment(s)?")

PostType = Literal["subreddit", "user"]


# When missing a '/' at the end of an url, reddit will make a redirection and it will reduce by 2 the number of requests remaining
def add_slash(url: str) -> str:
    path = url.split("/")

    if path[-1] != "" and not path[-1].startswith("?"):
        return url + "/"

    return url


def resolve_relative_url(path: str) -> str:
    return urljoin(OLD_REDDIT_BASE_URL, path)


def get_old_url(url: str) -> str:
    return url.replace("www.reddit", "old.reddit")


def get_new_url(url: str) -> str:
    return url.replace("old.reddit", "www.reddit")


def ensure_subreddit_url(name: str) -> str:
    if is_url(name):
        return name

    name = name.lstrip("/")

    if name.startswith("r/"):
        return urljoin(OLD_REDDIT_BASE_URL, name)

    return urljoin(OLD_REDDIT_BASE_URL + "/r/", name)


# TODO: make this a retryable method
def reddit_request(
    url: str, pool_manager: PoolManager
) -> Tuple[Response, WonderfulSoup, Optional[str]]:
    while True:
        response = request(
            add_slash(url),
            pool_manager=pool_manager,
            spoof_ua=True,
        )

        soup = response.soup()

        remaining_requests = float(response.headers["x-ratelimit-remaining"])

        if response.status == 429 and remaining_requests > 1:
            sleep(1)
            continue

        if (
            response.status == 500
            and soup.scrape_one("img", "alt") == "you broke reddit"
        ):
            return response, soup, "broken page"

        if response.status == 404 and soup.scrape_one("img", "alt") == "banned":
            return response, soup, "banned"

        if (
            soup.scrape_one("span.pagename.selected") == "page not found"
            or "search?q=" in url
            and soup.scrape_one("p.error")
        ):
            raise RedditInvalidTargetError

        if remaining_requests == 1 or response.status == 429:
            time_remaining = int(response.headers["x-ratelimit-reset"])
            sleep(time_remaining)
            continue

        return response, soup, None


def extract_t1_ids(text: str) -> List[str]:
    ids = [match.group(1) for match in re.finditer(ID_RE, text)]

    if ids:
        return ids

    return text.split("'")[-4].split(",")


def get_current_id(comment_element: Tag) -> str:
    current_id = comment_element.get("id")

    if current_id:
        current_id = current_id.split("_")[-1]
    else:
        current_id = comment_element["data-permalink"].split("/")[-2]

    return current_id


def get_points(element: Tag) -> str:
    scrapped_points = element.select_one(".score.unvoted")
    score_hidden = element.select_one(".score-hidden")

    if not scrapped_points and not score_hidden:
        return "deleted"

    scrapped_points = element.scrape_one(".score.unvoted", "title")

    if not scrapped_points:
        return "score hidden"

    return scrapped_points


def get_dates(element: Tag) -> Tuple[Optional[str], Optional[str]]:
    published_date = element.scrape_one("time", "datetime")
    edited_date = element.scrape_one("time.edited-timestamp", "datetime")

    return published_date, edited_date


def data_posts(
    post,
    title,
    url,
    author_text,
    points,
    raw_number_comments,
    number_comments,
    published_date,
    edited_date,
    link,
    error,
) -> RedditPost:
    author = post.scrape_one("a.author")

    if "reddit.com/" in link:
        link = None

    data = RedditPost(
        title=title,
        url=get_new_url(url),
        author=author if author else "[Deleted]",
        author_text=author_text,
        points=points,
        raw_number_comments=raw_number_comments,
        number_comments=number_comments,
        published_date=published_date,
        edited_date=edited_date,
        external_link=link,
        error=error,
    )

    return data


def data_user_posts(
    post,
    title,
    url,
    author_text,
    points,
    raw_number_comments,
    number_comments,
    published_date,
    edited_date,
    link,
    error,
) -> RedditUserPost:
    sub = post.scrape_one("a.subreddit", "href")

    data = RedditUserPost(
        title=title,
        url=get_new_url(url),
        author_text=author_text,
        points=points,
        raw_number_comments=raw_number_comments,
        number_comments=number_comments,
        published_date=published_date,
        edited_date=edited_date,
        external_link=link,
        subreddit=get_new_url(sub),
        error=error,
    )

    return data


class RedditScraper(object):
    def __init__(self):
        self.pool_manager = create_pool_manager()

    # NOTE: this potentially mutates the given list!
    def get_childs_l500(
        self,
        url: str,
        list_comments: List[Tuple[Optional[str], Tag]],
        parent_id: Optional[str],
    ) -> List[Tuple[Optional[str], Tag]]:
        _, soup, _ = reddit_request(url, self.pool_manager)

        comments = soup.select("div.commentarea > div > div.comment")

        if parent_id is None:
            for comment in comments:
                list_comments.append((None, comment))

            return list_comments

        for comment in comments:
            child = comment.force_select_one("div.child")

            if child.text != "":
                child = child.force_find("div")
                child_comment = child.find_all(
                    "div",
                    class_=lambda x: x
                    and (
                        "comment" in x
                        or "deleted comment" in x
                        or "morerecursion" in x
                        or "morechildren" in x
                    ),
                    recursive=False,
                )

                for element in child_comment:
                    list_comments.append((parent_id, element))

        return list_comments

    def get_comments(self, url: str, all: bool = True) -> Iterator[RedditComment]:
        m_comments: List[Tuple[Optional[str], Tag]] = []

        old_url = get_old_url(url)
        url_limit = old_url + "?limit=500"

        _, soup, error = reddit_request(url_limit, self.pool_manager)

        # TODO: this should be an exception?
        if error:
            yield RedditComment(
                comment_url="",
                author="",
                id="",
                parent="",
                points="",
                published_date="",
                edited_date="",
                comment="",
                error=error,
            )

            return

        first_comments = soup.select("div.commentarea > div > div.comment")

        if all:
            more = soup.select("div.commentarea > div > div.morechildren")

            for element in more:
                a = element.force_select_one("a")
                onclick = a["onclick"]
                id_list = extract_t1_ids(onclick)

                for id in id_list:
                    comment_url = f"{old_url}{id}/"
                    m_comments = self.get_childs_l500(comment_url, m_comments, None)

        for element in first_comments:
            m_comments.append((None, element))

        while m_comments:
            parent, comment = m_comments.pop()
            current_id = get_current_id(comment)

            if "deleted" in comment["class"] and "comment" in comment["class"]:
                comment_url = comment["data-permalink"]
                author = "[Deleted]"
                points = None
            else:
                comment_url = comment.force_scrape_one("a.bylink", "href")
                author = comment.scrape_one("div.entry.unvoted a.author")
                if not author:
                    author = "[Deleted]"
                points = get_points(comment)

            published_date, edited_date = get_dates(comment)

            if "morerecursion" in comment["class"] and all:
                url_rec = f"https://old.reddit.com{comment.scrape_one('a', 'href')}"
                m_comments = self.get_childs_l500(url_rec, m_comments, parent)
            elif "morechildren" in comment["class"] and all:
                a = comment.force_select_one("a")
                onclick = a["onclick"]
                id_list = extract_t1_ids(onclick)
                for id in id_list:
                    comment_url = f"{old_url}{id}/"
                    m_comments = self.get_childs_l500(
                        comment_url, m_comments, current_id
                    )
            else:
                child = comment.force_select_one("div.child")

                if child.text != "":
                    child = child.force_find("div")
                    if all:
                        child_comment = child.find_all(
                            "div",
                            class_=lambda x: x
                            and (
                                "comment" in x
                                or "morerecursion" in x
                                or "morechildren" in x
                            ),
                            recursive=False,
                        )
                    else:
                        child_comment = child.find_all(
                            "div",
                            class_=lambda x: x and ("comment" in x),
                            recursive=False,
                        )
                    for element in child_comment:
                        m_comments.append((current_id, element))

                data = RedditComment(
                    comment_url=get_new_url(resolve_relative_url(comment_url)),
                    author=author if author else "[Deleted]",
                    id=current_id,
                    parent=parent,
                    points=points,
                    published_date=published_date,
                    edited_date=edited_date,
                    comment=comment.force_scrape_one("div.md:not(div.child a)"),
                    error=error,
                )

                if data.id != "":
                    yield data

    @overload
    def get_posts(
        self, url: str, type: Literal["subreddit"], add_text: bool, limit: int
    ) -> Iterator[RedditPost]: ...

    @overload
    def get_posts(
        self, url: str, type: Literal["user"], add_text: bool, limit: int
    ) -> Iterator[RedditUserPost]: ...

    def get_posts(
        self, url: str, type: PostType, add_text: bool, limit: int
    ) -> Iterator[Union[RedditPost, RedditUserPost]]:
        fn = data_posts if type == "subreddit" else data_user_posts
        n_crawled = 0
        old_url = get_old_url(ensure_subreddit_url(url))
        while old_url and (limit is None or n_crawled < limit):
            if limit is not None and n_crawled == limit:
                break
            _, soup, error = reddit_request(old_url, self.pool_manager)
            posts = soup.select("div[id^='thing_t3_']")
            for post in posts:
                if limit is not None and n_crawled == limit:
                    break
                list_buttons = post.force_select_one("ul.flat-list.buttons")
                if len(list_buttons.scrape("span.promoted-span")) == 0:
                    title = post.force_select_one("a.title").get_text()
                    post_url = list_buttons.force_scrape_one(
                        "a.bylink.comments", "href"
                    )
                    n_comments_scraped = list_buttons.force_select_one(
                        "a.bylink.comments"
                    ).get_text()
                    match = COMMENTS_NUMBER_RE.match(n_comments_scraped)
                    if match:
                        n_comments = int(match.group(1))
                    else:
                        n_comments = 0
                    upvote = get_points(post)
                    published_date, edited_date = get_dates(post)
                    link = resolve_relative_url(
                        post.force_scrape_one("a.title", "href")
                    )
                    if link == post_url:
                        link = ""
                    if add_text:
                        _, text_soup, text_error = reddit_request(
                            post_url, self.pool_manager
                        )
                        if text_error:
                            yield fn(
                                post,
                                title,
                                post_url,
                                None,
                                upvote,
                                n_comments_scraped,
                                n_comments,
                                published_date,
                                edited_date,
                                link,
                                text_error,
                            )
                        content = text_soup.scrape_one(
                            "div#siteTable div.usertext-body"
                        )
                    else:
                        content = ""
                    yield fn(
                        post,
                        title,
                        post_url,
                        content,
                        upvote,
                        n_comments_scraped,
                        n_comments,
                        published_date,
                        edited_date,
                        link,
                        error,
                    )
                    n_crawled += 1
            old_url = soup.scrape_one("span.next-button a", "href")

    def get_user_comments(self, url: str, limit: int) -> Iterator[RedditUserComment]:
        n_crawled = 0
        old_url = get_old_url(url)
        while old_url and (limit is None or n_crawled < limit):
            if limit is not None and n_crawled == limit:
                break
            _, soup, error = reddit_request(old_url, self.pool_manager)
            if error:
                yield RedditUserComment(
                    post_title="",
                    post_url="",
                    link="",
                    post_author="",
                    post_subreddit="",
                    points="",
                    published_date="",
                    edited_date="",
                    text="",
                    comment_url="",
                    error=error,
                )
            else:
                comments = soup.select("[data-type='comment']")
                for comment in comments:
                    if limit is not None and n_crawled == limit:
                        break
                    post_title = comment.force_scrape_one("a.title")
                    post_url = comment.force_scrape_one("a.bylink.may-blank", "href")
                    post_author = comment.force_scrape_one("p.parent>a.author")
                    post_subreddit = comment.force_scrape_one("a.subreddit", "href")
                    points = get_points(comment)
                    published_date, edited_date = get_dates(comment)
                    text = comment.force_scrape_one("div.content div.md")
                    link = comment.force_scrape_one("div.content div.md a", "href")
                    comment_url = comment.force_scrape_one("a.bylink", "href")
                    data = RedditUserComment(
                        post_title=post_title,
                        post_url=get_new_url(post_url),
                        post_author=post_author,
                        post_subreddit=get_new_url(post_subreddit),
                        points=points,
                        published_date=published_date,
                        edited_date=edited_date,
                        text=text,
                        link=link,
                        comment_url=get_new_url(comment_url),
                        error=error,
                    )
                    yield data
                    n_crawled += 1
            old_url = soup.scrape_one("span.next-button a", "href")



from urllib.parse import urljoin
from websockets.sync.client import connect, ClientConnection

from minet.bluesky.constants import BLUESKY_FIREHOSE_BASE_URL

# TODO: investigate libipld


class BlueskyWebSocketClient:
    def subscribe_repos(self) -> ClientConnection:
        return connect(
            urljoin(BLUESKY_FIREHOSE_BASE_URL, "com.atproto.sync.subscribeRepos")
        )



from typing import Optional

from ural import URLFormatter

from minet.bluesky.constants import BLUESKY_HTTP_API_BASE_URL


class BlueskyHTTPAPIUrlFormatter(URLFormatter):
    BASE_URL = BLUESKY_HTTP_API_BASE_URL

    def create_session(self) -> str:
        return self.format(
            path="com.atproto.server.createSession",
        )

    def refresh_session(self) -> str:
        return self.format(path="com.atproto.server.refreshSession")

    def search_posts(
        self, q: str, cursor: Optional[str] = None, limit: int = 100
    ) -> str:
        return self.format(
            path="app.bsky.feed.searchPosts",
            args={"q": q, "cursor": cursor, "limit": limit},
        )



# =============================================================================
# Minet Reddit Posts CLI Action
# =============================================================================
#
# Logic of the `rd posts` action.
#
from minet.cli.utils import with_enricher_and_loading_bar

from minet.reddit.scraper import RedditScraper
from minet.reddit.types import RedditPost
from minet.reddit.exceptions import RedditInvalidTargetError


@with_enricher_and_loading_bar(
    headers=RedditPost,
    title="Scraping posts",
    unit="pages",
    nested=True,
    sub_unit="posts",
)
def action(cli_args, enricher, loading_bar):
    scraper = RedditScraper()

    for i, row, url in enricher.enumerate_cells(
        cli_args.column, with_rows=True, start=1
    ):
        with loading_bar.step(url):
            try:
                posts = scraper.get_posts(
                    url, "subreddit", cli_args.text, cli_args.limit
                )

            except RedditInvalidTargetError:
                loading_bar.print(
                    "the script could not complete normally on line %i" % (i)
                )
                continue

            for post in posts:
                loading_bar.nested_advance()
                enricher.writerow(row, post)



from typing import Iterator, Optional

from time import time, sleep

from minet.web import (
    create_request_retryer,
    create_pool_manager,
    retrying_method,
    request,
    Response,
)

from minet.bluesky.urls import BlueskyHTTPAPIUrlFormatter
from minet.bluesky.jwt import parse_jwt_for_expiration
from minet.bluesky.types import BlueskyPost
from minet.bluesky.exceptions import (
    BlueskyAuthenticationError,
    BlueskySessionRefreshError,
)


class BlueskyHTTPClient:
    def __init__(self, identifier: str, password: str):
        self.urls = BlueskyHTTPAPIUrlFormatter()
        self.pool_manager = create_pool_manager()
        self.retryer = create_request_retryer()
        self.rate_limit_reset: Optional[int] = None

        # First auth
        self.create_session(identifier, password)

    @retrying_method()
    def create_session(self, identifier: str, password: str):
        response = request(
            self.urls.create_session(),
            pool_manager=self.pool_manager,
            method="POST",
            json_body={"identifier": identifier, "password": password},
        )

        if response.status != 200:
            raise BlueskyAuthenticationError

        data = response.json()

        self.access_jwt = data["accessJwt"]
        self.refresh_jwt = data["refreshJwt"]
        self.access_jwt_expiration = parse_jwt_for_expiration(self.access_jwt)

    def is_access_jwt_expired(self) -> bool:
        # If the token has 10 seconds left, we consider it expired to avoid network-related issues
        return self.access_jwt_expiration - time() < 10

    @retrying_method()
    def refresh_session(self):
        response = request(
            self.urls.refresh_session(),
            pool_manager=self.pool_manager,
            method="POST",
            headers={"Authorization": "Bearer {}".format(self.refresh_jwt)},
        )

        if response.status != 200:
            raise BlueskySessionRefreshError

        data = response.json()

        self.access_jwt = data["accessJwt"]
        self.refresh_jwt = data["refreshJwt"]
        self.access_jwt_expiration = parse_jwt_for_expiration(self.access_jwt)

    @retrying_method()
    def request(
        self,
        url: str,
        method: str = "GET",
        json_body=None,
    ) -> Response:
        if self.rate_limit_reset is not None:
            sleep(self.rate_limit_reset - time() + 0.1)
            self.rate_limit_reset = None

        if self.is_access_jwt_expired():
            self.refresh_session()

        headers = {"Authorization": "Bearer {}".format(self.access_jwt)}

        response = request(
            url,
            pool_manager=self.pool_manager,
            method=method,
            json_body=json_body,
            known_encoding="utf-8",
            headers=headers,
        )

        remaining = int(response.headers["RateLimit-Remaining"])

        if remaining <= 0:
            self.rate_limit_reset = int(response.headers["RateLimit-Reset"])

        return response

    def search_posts(self, query: str) -> Iterator[BlueskyPost]:
        cursor = None

        while True:
            url = self.urls.search_posts(query, cursor=cursor)

            response = self.request(url)
            data = response.json()

            for post in data["posts"]:
                yield BlueskyPost.from_payload(post)

            cursor = data.get("cursor")

            if cursor is None:
                break



from minet.exceptions import MinetError


class BlueskyError(MinetError):
    pass


class BlueskyAuthenticationError(BlueskyError):
    pass


class BlueskySessionRefreshError(BlueskyError):
    pass



from typing import Optional, List

from dataclasses import dataclass
from casanova import TabularRecord

# TODO: embeds, mentions, tags, links, reply, quote


@dataclass
class BlueskyPost(TabularRecord):
    uri: str
    author_handle: str
    author_did: str
    author_display_name: Optional[str]
    created_at: str
    text: str
    reply_count: int
    repost_count: int
    like_count: int
    quote_count: int
    langs: List[str]

    @classmethod
    def from_payload(cls, payload) -> "BlueskyPost":
        author = payload["author"]
        record = payload["record"]

        return cls(
            uri=payload["uri"],
            author_handle=author["handle"],
            author_did=author["did"],
            author_display_name=author.get("displayName"),
            created_at=record["createdAt"],
            text=record["text"],
            reply_count=payload["replyCount"],
            repost_count=payload["repostCount"],
            like_count=payload["likeCount"],
            quote_count=payload["quoteCount"],
            langs=record.get("langs", []),
        )



# =============================================================================
# Minet Reddit Comments CLI Action
# =============================================================================
#
# Logic of the `rd user_comments` action.
#
from minet.cli.utils import with_enricher_and_loading_bar

from minet.reddit.scraper import RedditScraper
from minet.reddit.types import RedditUserComment
from minet.reddit.exceptions import RedditInvalidTargetError


@with_enricher_and_loading_bar(
    headers=RedditUserComment,
    title="Scraping user comments",
    unit="pages",
    nested=True,
    sub_unit="comments",
)
def action(cli_args, enricher, loading_bar):
    scraper = RedditScraper()

    for i, row, url in enricher.enumerate_cells(
        cli_args.column, with_rows=True, start=1
    ):
        with loading_bar.step(url):
            try:
                posts = scraper.get_user_comments(url, cli_args.limit)

            except RedditInvalidTargetError:
                loading_bar.print(
                    "the script could not complete normally on line %i" % (i)
                )
                continue

            for post in posts:
                loading_bar.nested_advance()
                enricher.writerow(row, post)



from casanova import Enricher

from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.loading_bar import LoadingBar

from minet.cli.bluesky.utils import with_bluesky_fatal_errors

from minet.bluesky import BlueskyHTTPClient
from minet.bluesky.types import BlueskyPost


@with_bluesky_fatal_errors
@with_enricher_and_loading_bar(
    headers=BlueskyPost,
    title="Searching posts",
    unit="queries",
    nested=True,
    sub_unit="posts",
)
def action(cli_args, enricher: Enricher, loading_bar: LoadingBar):
    client = BlueskyHTTPClient(cli_args.identifier, cli_args.password)

    for row, query in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step(query):
            for post in client.search_posts(query):
                enricher.writerow(row, post)
                loading_bar.nested_advance()
